/* test_libFLAC - Unit tester for libFLAC
 * Copyright (C) 2014  Xiph.Org Foundation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>

#include "FLAC/assert.h"
#include "share/compat.h"
#include "private/md5.h"
#include "md5.h"


static FLAC__bool test_md5_clear_context(void);
static FLAC__bool test_md5_codec(void);
static FLAC__bool test_md5_accumulate(const FLAC__int32 * const * signal,unsigned channels, unsigned samples, unsigned bytes_per_sample, const FLAC__byte target_digest [16]);

FLAC__bool test_md5(void)
{
	printf("\n+++ libFLAC unit test: md5\n\n");

	if (! test_md5_clear_context())
		return false;

	if (! test_md5_codec())
		return false;

	printf("\nPASSED!\n");
	return true;
}

/*----------------------------------------------------------------------------*/

static FLAC__bool test_md5_clear_context(void)
{
	FLAC__MD5Context ctx;
	FLAC__byte digest[16];
	unsigned k ;
	char * cptr;

	printf("testing FLAC__MD5Init ... ");
	FLAC__MD5Init (&ctx);
	if (ctx.buf[0] != 0x67452301) {
		printf("FAILED!\n");
		return false;
	}
	printf("OK\n");

	printf("testing that FLAC__MD5Final clears the MD5Context ... ");
	FLAC__MD5Final(digest, &ctx);
	cptr = (char*) &ctx ;
	for (k = 0 ; k < sizeof (ctx) ; k++) {
		if (cptr [k]) {
			printf("FAILED, MD5 ctx has not been cleared after FLAC__MD5Final\n");
			return false;
		}
	}

	printf("OK\n");
	return true;
}

#define ARRAY_LEN(x)   (sizeof (x) / sizeof (x [0]))

static FLAC__byte target_digests [8][4][16] =
{	/* 1 channel */
	{ 	/* 1 byte per sample */
		{ 0x41, 0x91, 0x23, 0x31, 0x5a, 0x22, 0x59, 0x2e, 0x31, 0x70, 0xb6, 0x7d, 0x0a, 0xa2, 0x69, 0x16 },
		/* 2 bytes per sample */
		{ 0xab, 0x7d, 0x77, 0x99, 0xdc, 0x15, 0x4a, 0xbc, 0x37, 0x95, 0xe5, 0x26, 0xce, 0x3c, 0x3f, 0x25 },
		/* 3 bytes per sample */
		{ 0x03, 0x92, 0xa7, 0x6d, 0x78, 0xa8, 0x8a, 0x2b, 0x7c, 0x80, 0x8e, 0x6c, 0x09, 0xf6, 0x8b, 0x73 },
		/* 4 bytes per sample */
		{ 0xbd, 0x04, 0x81, 0x60, 0xee, 0x96, 0xb8, 0x42, 0x18, 0x65, 0x9d, 0x40, 0x21, 0x67, 0x03, 0x87 }
	},

	/* 2 channels */
	{
		{ 0xae, 0x46, 0xa7, 0x20, 0x76, 0xf2, 0xa3, 0x76, 0xe0, 0xaf, 0xc4, 0x33, 0x5a, 0x90, 0x6d, 0x38 },
		{ 0x52, 0x43, 0x94, 0xa0, 0xe8, 0xa8, 0xa8, 0x2e, 0xfe, 0x90, 0x1c, 0x0a, 0x34, 0x89, 0xb2, 0x40 },
		{ 0xf4, 0x9f, 0xa5, 0xf0, 0x8d, 0x4e, 0x3d, 0x68, 0x36, 0xdc, 0x43, 0x9a, 0x88, 0x84, 0xba, 0xd0 },
		{ 0xb1, 0xcf, 0xe2, 0x11, 0x45, 0x57, 0xe8, 0x98, 0xe3, 0xdc, 0x7b, 0xcd, 0x4e, 0xe3, 0xde, 0x86 }
	},

	/* 3 channels */
	{
		{ 0x4e, 0x55, 0x01, 0x00, 0x03, 0x9d, 0x90, 0xd9, 0x44, 0xbd, 0x46, 0x01, 0x14, 0xd4, 0x66, 0x90 },
		{ 0xa7, 0xa5, 0xfe, 0x33, 0x20, 0x4c, 0xf3, 0x0c, 0x13, 0x8e, 0x65, 0x08, 0xce, 0xe4, 0xdb, 0xc7 },
		{ 0xfa, 0xc8, 0xe2, 0x4a, 0xf4, 0x9e, 0xe8, 0x20, 0x46, 0x4c, 0xaa, 0x9f, 0xe2, 0x02, 0x36, 0x6f },
		{ 0x4f, 0xa9, 0x19, 0xf2, 0x6d, 0x6b, 0xbf, 0x46, 0x6b, 0xfe, 0xec, 0xc9, 0xb7, 0x37, 0xcd, 0x3e }
	},

	/* 4 channels */
	{
		{ 0xb0, 0x45, 0xd1, 0x25, 0x13, 0x57, 0x3a, 0xd6, 0x5b, 0x0e, 0x7c, 0x04, 0x41, 0xcb, 0x20, 0xc2 },
		{ 0x14, 0xc5, 0xce, 0xbc, 0xdc, 0xb1, 0xd2, 0x36, 0xce, 0xc2, 0x05, 0x5f, 0x0d, 0x85, 0xfe, 0x08 },
		{ 0x12, 0xd3, 0xb2, 0x43, 0x14, 0x8c, 0x8a, 0x46, 0x79, 0xc9, 0x33, 0x50, 0x22, 0xef, 0xa9, 0xa9 },
		{ 0xed, 0xa8, 0x21, 0xf9, 0x7d, 0xbe, 0x24, 0x57, 0xb3, 0x99, 0x6a, 0x01, 0x27, 0x79, 0x70, 0xd6 }
	},

	/* 5 channels */
	{
		{ 0x30, 0x1b, 0xbc, 0xe4, 0x75, 0x4c, 0xe8, 0x51, 0x6a, 0x72, 0x04, 0xf3, 0xbd, 0xcb, 0x56, 0x9f },
		{ 0x89, 0x92, 0xa0, 0x1f, 0x76, 0xfe, 0x3b, 0x10, 0x1e, 0xea, 0xcc, 0x66, 0x83, 0xa1, 0x5f, 0xd2 },
		{ 0xb7, 0x49, 0xed, 0xe0, 0x30, 0xef, 0x7b, 0x16, 0x88, 0xd1, 0x09, 0x9b, 0x9a, 0x18, 0xba, 0x1d },
		{ 0x43, 0x9a, 0x2f, 0xf2, 0x10, 0x44, 0x4e, 0x0d, 0xbf, 0x8c, 0x19, 0xf6, 0x6e, 0x4b, 0xa1, 0x00 }
	},

	/* 6 channels */
	{
		{ 0xfd, 0x27, 0x4f, 0x12, 0xd2, 0x99, 0xee, 0x3a, 0xf0, 0x5a, 0xd8, 0x87, 0x23, 0x6f, 0xcf, 0x3b },
		{ 0xbd, 0x50, 0x93, 0x17, 0x9d, 0xd4, 0x16, 0x18, 0xe2, 0xbd, 0x10, 0x96, 0x3f, 0x8d, 0xa0, 0x63 },
		{ 0xbf, 0xc0, 0xad, 0xfc, 0x53, 0x4d, 0xca, 0x25, 0xa1, 0x1c, 0x37, 0x40, 0xa7, 0x0e, 0xd2, 0x37 },
		{ 0x54, 0x38, 0x05, 0x54, 0x7f, 0xf8, 0x1a, 0xf8, 0x58, 0xea, 0x2e, 0x3b, 0x03, 0x1e, 0xc9, 0x24 }
	},

	/* 7 channels */
	{
		{ 0x45, 0x1d, 0x11, 0x6f, 0xd2, 0x0a, 0xfb, 0xe5, 0xd0, 0x83, 0x10, 0x0f, 0xe0, 0x44, 0x45, 0xd2 },
		{ 0x0b, 0x6f, 0x97, 0xd1, 0x9f, 0x4a, 0xf5, 0xa6, 0xab, 0x17, 0x2a, 0xd8, 0xd3, 0x24, 0x43, 0x68 },
		{ 0x63, 0x93, 0xe1, 0x89, 0xf5, 0xc8, 0x58, 0xac, 0x1b, 0x87, 0xb0, 0x8f, 0x59, 0x38, 0xd0, 0xf4 },
		{ 0xc6, 0x81, 0x83, 0xde, 0x10, 0xf2, 0x94, 0x00, 0x31, 0xdd, 0x8e, 0x4b, 0x0f, 0x5a, 0x95, 0x0b }
	},

	/* 8 channels */
	{
		{ 0xf5, 0x4b, 0xea, 0xa6, 0x2a, 0x2b, 0x32, 0x66, 0xa7, 0xa8, 0xf8, 0x0e, 0x56, 0x07, 0xe0, 0x23 },
		{ 0x6e, 0xd4, 0xbc, 0x21, 0x7d, 0x07, 0x02, 0x89, 0x80, 0xe8, 0xe6, 0x39, 0x73, 0x18, 0xc2, 0xa0 },
		{ 0x6a, 0x28, 0xa1, 0xd8, 0x7c, 0x2f, 0xd9, 0x79, 0x35, 0xa0, 0x6c, 0x53, 0x81, 0x7e, 0x40, 0x9b },
		{ 0xdf, 0x4e, 0xe2, 0x37, 0x88, 0x3d, 0xfd, 0x94, 0xf4, 0x47, 0x00, 0x76, 0x55, 0x6f, 0xea, 0xfd }
	}
};

static FLAC__bool test_md5_codec(void)
{
	FLAC__int32 arrays[8][64], *pointer[8], **signal;
	FLAC__int32 seed = 0x12345679;
	unsigned chan, byte_size;

	/* Set up signal data using a trival Linear Congruent PRNG. */
	signal = &pointer[0];
	for (chan = 0 ; chan < ARRAY_LEN (arrays) ; chan ++) {
		unsigned k;
		pointer[chan] = arrays [chan];
		for (k = 0 ; k < ARRAY_LEN (arrays [0]) ; k++) {
			seed = seed * 1103515245 + 12345;
			arrays[chan][k] = seed;
		}
	}

	for (chan = 1 ; chan <= ARRAY_LEN (arrays) ; chan ++) {
		for (byte_size = 1 ; byte_size <= 4 ; byte_size ++) {
			if (! test_md5_accumulate((const FLAC__int32 * const *) signal, chan, ARRAY_LEN (signal [0]), byte_size, target_digests[chan-1][byte_size-1]))
				return false;
		}
	}

	return true;
}

static FLAC__bool test_md5_accumulate(const FLAC__int32 * const * signal, unsigned channels, unsigned samples, unsigned bytes_per_sample, const FLAC__byte target_digest [16])
{
	FLAC__MD5Context ctx;
	FLAC__byte digest[16];

	memset(&ctx, 0, sizeof (ctx));

	printf("testing FLAC__MD5Accumulate (channels=%d, bytes_per_sample=%d) ... ", channels, bytes_per_sample);

	FLAC__MD5Init(&ctx);
	FLAC__MD5Accumulate(&ctx, signal, channels, samples, bytes_per_sample);
	FLAC__MD5Final(digest, &ctx);

	if (memcmp(digest, target_digest, sizeof (digest))) {
		int k ;

		printf("\nFAILED, expected MD5 sum ");
		for (k = 0 ; k < 16 ; k++)
			printf("%02x", (digest [k] & 0xff));
		printf (" but got ");
		for (k = 0 ; k < 16 ; k++)
			printf("%02x", (target_digest [k] & 0xff));
		puts("\n");
		return false;
	}

	printf("OK\n");
	return true;
}
